/* ds.cxx
     $Id: ds.cxx,v 1.9 1998/12/14 09:08:17 elf Exp $

   written by Oscar Levi
   6 December 1998

   This file is part of the project BSIGN.  See the file README for
   more information.

   Copyright (c) 1998 The Buici Company.

   This program is free software; you can redistribute it and/or modify 
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

   -----------
   DESCRIPTION
   -----------

   Interface for generating digital signature certificates.  

   The first round uses gpg for generating signatures.

*/


#include "standard.h"
#include "exec.h"
#include "ds.h"
#include "exitstatus.h"
#include "tty.h"

extern bool g_fDebug;
extern char* g_szApplication;

const char* fetch_passphrase (void);

/* create_digital_signature

   takes the cb byte data block at pb, writes it to a temporary file
   invokes the appropriate software to sign it, and returns the data
   in a memory block.  The first two bytes of the data block are the
   MSB ordered length of the signature.

   **FIXME: we want to prevent this going to disk because someone
   could be very clever and intercept it before we can write the
   signature.

*/

char* create_digital_signature (const char* pb, size_t cb)
{
				// Create control streams
  char rgbSignature[512];
  char rgbErr[1024];		// *** FIXME: we need a ring buffer maybe?
  const char* szPass = fetch_passphrase ();
  ExStream exsIn, exsOut, exsPassPhrase, exsErr;
  if (   !exsIn.for_stdin (pb, cb)
      || !exsOut.for_stdout (rgbSignature, sizeof (rgbSignature))
      || !exsPassPhrase.for_writing (szPass, strlen (szPass))
      || !exsErr.for_stderr (rgbErr, sizeof (rgbErr))
      )
    return NULL;

  char sz[256];
  sprintf (sz, "gpg --no-greeting -sb -o - --passphrase-fd %d", 
	   exsPassPhrase.fd_write ());

  ExStream* rgexs[] = { &exsIn, &exsOut, &exsPassPhrase, &exsErr };
  int result = ExStream::exec (sz, rgexs, sizeof (rgexs)/sizeof (ExStream*));

  cb = exsOut.used ();
  char* pbReturn = (char*) malloc (cb + 2);
  pbReturn[0] = ((cb >> 8) & 0xff);
  pbReturn[1] = (cb & 0xff);
  memcpy (pbReturn + 2, rgbSignature, cb);

  return pbReturn;
}

eExitStatus verify_digital_signature (const char* pbData, size_t cbData,
				      const char* pbCert, size_t cbCert)
{
  if (cbCert == 0)		// Kind of a sanity check.  We also
    return badsignature;	// verify this at the next layer up.
				// This is here in case we call from
				// some other place.

				// Create control streams
  char rgbErr[1024];
  ExStream exsIn, exsErr;
  if (   !exsIn.for_stdin (pbCert, cbCert)
      || !exsErr.for_stderr (rgbErr, sizeof (rgbErr)))
    return toomanyopenfiles;

#if 1
  char szTemplate[30];
  strcpy (szTemplate, "/tmp/XXXXXX");
  int fh = mkstemp (szTemplate);
				// *** FIXME, more error handling?
  if (g_fDebug)
    printf ("%s: creating temporary '%s' for data on verify\n", 
	    g_szApplication, szTemplate);

  write (fh, pbData, cbData);
  close (fh);
#endif

  char sz[256];
  sprintf (sz, "gpg --no-greeting --verify - %s", szTemplate);
  ExStream* rgexs[] = { &exsIn, &exsErr };
  int result = ExStream::exec (sz, rgexs, sizeof (rgexs)/sizeof (ExStream*));
  unlink (szTemplate);

  //  fprintf (stderr, "exec result is %d\n", result);
  return result ? badsignature : noerror;
}


/* fetch_passphrase

   ask the user for the passphrase.  We cache this phrase and return
   it for every subsequenct call.  This is fine, for now.  It is
   possible that we'll do something more sophisticated later, but it
   is probably not necessary.

*/

const char* fetch_passphrase (void)
{
  static char* g_szPassPhrase;

  if (g_szPassPhrase)
    return g_szPassPhrase;

  g_szPassPhrase = (char*) malloc (512);

				// *** FIXME make this raw or
				// something so we don't ECHO.  We do
				// want to continue to input until we
				// receive a newline, though.

  disable_echo (0);
  printf ("\nEnter pass phrase: ");
  fflush (stdout);
  size_t cb = read (0, g_szPassPhrase, 511);
  restore_echo (0);
  g_szPassPhrase[cb] = 0;

  return g_szPassPhrase;
}
