/* main.cxx
     $Id: main.cxx,v 1.13 1998/12/14 09:08:17 elf Exp $

   written by Oscar Levi
   1 December 1998
   
   This file is part of the project BSIGN.  See the file README for
   more information.

   Copyright (c) 1998 The Buici Company.

   This program is free software; you can redistribute it and/or modify 
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

*/

  /* ----- Includes */

#include "standard.h"

#define DECLARE_VERSION
#include "version.h"

#include <sys/mman.h>
#include <sys/stat.h>

#include "bsign.h"
#include "options.h"
#include "files.h"
#include "exitstatus.h"

extern "C" void sha_memory(void*, unsigned32, unsigned32*);


  /* ----- Class Globals/Statics */

int g_fPalette;
int g_fSymbols;
int g_fSections;
int g_fDetail;

int g_fSign;			// !0: hash & sign files 
int g_fHash;			// !0: hash files instead, 0: check signature
bool g_fExpectSignature;	// !0: check for signature as well as hash
bool g_fProcessedFile;		// Set when we do something
eExitStatus g_exitStatus;	// Return code for the program
int g_fQuiet;			// Inhibit messages, return code only
int g_fDebug;			// Show debug messages
char* g_szOutput;		// Name of output file

char* g_szApplication;		// Name of application

int do_filename (OPTION* pOption, const char* pch);
int do_unrecognized (OPTION*, const char*);
int do_usage (OPTION*, const char*);
int do_version (OPTION*, const char*);
int do_hash (OPTION*, const char*);
int do_sign (OPTION*, const char*);
int do_check (OPTION*, const char*);
int do_verify (OPTION*, const char*);

OPTION rgOptions[] =
{
  { "sign",		0, NULL, do_sign  },
  { "s",		0, NULL, do_sign  },

  { "hash",		0, NULL, do_hash  },
  { "H",		0, NULL, do_hash  },

  { "quiet",		OPTION_F_SET_INT,		      &g_fQuiet },
  { "q",		OPTION_F_SET_INT,		      &g_fQuiet },

  { "debug",		OPTION_F_SET_INT,		      &g_fDebug },
  { "d",		OPTION_F_SET_INT,		      &g_fDebug },

  { "check",		0, NULL, do_check  },
  { "c",		0, NULL, do_check  },

  { "verify",		0, NULL, do_verify  },
  { "v",		0, NULL, do_verify  },

  { "output",		OPTION_F_SET_STRING,		    &g_szOutput },
  { "o",		OPTION_F_SET_STRING,		    &g_szOutput },

  { "help",		0, NULL, do_usage				},
  { "h",		0, NULL, do_usage				},

  { "version",		0, NULL, do_version				},
  { "V",		0, NULL, do_version				},

  { "",			OPTION_F_NONOPTION, NULL, do_filename		},
  { "",			OPTION_F_DEFAULT, NULL,	do_unrecognized		},
  { NULL								},
};

  /* ----- Methods */

int do_check (OPTION* /* pOption */, const char* /* pch */)
{
  g_fSign = g_fHash = false;
  g_fExpectSignature = true;
  return 0;
}

int do_verify (OPTION* /* pOption */, const char* /* pch */)
{
  g_fSign = g_fHash = false;
  g_fExpectSignature = true;
  return 0;
}

int do_sign (OPTION* /* pOption */, const char* /* pch */)
{
  g_fSign = true;
  g_fHash = true;
  return 0;
}

int do_hash (OPTION* /* pOption */, const char* /* pch */)
{
  g_fSign = false;
  g_fHash = true;
  return 0;
}

int do_filename (OPTION* /* pOption */, const char* pch)
{
  int fh = open (pch, O_RDONLY);

   g_fProcessedFile = true;

   //   printf ("do_filename: '%s'\n", pch);

   if (fh == -1) {
     fprintf (stderr, "%s: file not found '%s'\n", g_szApplication, pch);
     g_exitStatus = filenotfound;
     return 1;			// OK, no resources allocated
   }
   
   char* szFileNew = NULL;
   int fhNew = 0;
   if (g_fSign | g_fHash) {
     char* szPath = path_of (g_szOutput ? g_szOutput : pch);
     szFileNew = new char [strlen (szPath) + 32];
     strcpy (szFileNew, szPath);
     strcat (szFileNew, "/bsign.XXXXXX");
     fhNew = mkstemp (szFileNew);
     if (fhNew == -1) {
       free (szPath);
       szPath = path_of (g_szOutput ? g_szOutput : pch);
       fprintf (stderr, "%s: unable to create file in '%s'\n", 
		g_szApplication, szPath);
       close (fh);
       free (szPath);
       delete szFileNew;
       g_exitStatus = permissiondenied;
       return 1;		// OK, all strings freed
     }
     //     printf ("opened '%s'\n", szFileNew);
   }

   char* pbMap = NULL;
   struct stat stat;
   long cb;
   if (!::stat (pch, &stat) && (cb = stat.st_size)
      && (pbMap = (char*) mmap (NULL, cb, PROT_READ,
                                MAP_FILE | MAP_PRIVATE, fh, 0))) {
     if (g_fSign | g_fHash) {
       if (is_elf (pbMap, cb)) 
	 hash_elf (pbMap, cb, fhNew, g_fSign);
       else {
	 printf ("file '%s' is not ELF\n", pch);
	 g_exitStatus = unsupportedfiletype;
       }
     }
     else {
       if (!is_elf (pbMap, cb)) {
	 printf ("file '%s' is not ELF\n", pch);
	 g_exitStatus = unsupportedfiletype;
       }
       else {
	 g_exitStatus = check_elf (pbMap, cb, g_fExpectSignature);
	 if (!g_fQuiet) {
	   switch (g_exitStatus) {
	   case noerror:
	     printf ("Good %s found in", 
		     g_fExpectSignature ? "signature" : "hash");
	     break;
	   case badhash:
	     printf ("Invalid hash found in");
	     break;
	   case badsignature:
	     printf ("Invalid signature found in");
	     break;
	   case nohash:
	     printf ("No hash found in");
	     break;
	   case nosignature:
	     printf ("No signature found in");
	     break;
	   default:
	     printf ("error %d while verifying", g_exitStatus);
	     break;
	   }
	   printf (" '%s'.\n", pch);
	 }
       }
     }
   }

   if (pbMap)
     munmap (pbMap, cb);
   if (fhNew) {
     close (fhNew);
     if (!g_exitStatus 
	 && !dup_status (szFileNew, pch)) {
       unlink (szFileNew);
       fprintf (stderr, "%s: insufficient priviledge to duplicate '%s'"
		" ownership\n", g_szApplication, pch);
       g_exitStatus = permissiondenied;
     }
     if (!g_exitStatus 
	 && !replace_file (g_szOutput ? g_szOutput : pch, szFileNew)) {
       fprintf (stderr, "%s: unable to create output file '%s'\n", 
		g_szApplication, g_szOutput ? g_szOutput : pch);
				// I'd be suprised if I ever get here.
				// I think that by this time, I have
				// already written the file in the
				// same directory and performed chown
				// to set the permissions.
       g_exitStatus = permissiondenied;
     }
   }
   if (szFileNew)
     delete szFileNew;
   close (fh);
   return g_exitStatus ? 1 : 0;
}


int do_unrecognized (OPTION* /* pOption */, const char* pch)
{
  printf ("%s: unrecognized option '%s'\n",
	  g_szApplication, pch);
  printf ("Try '%s --help' for usage information.\n",
	  g_szApplication);
  g_exitStatus = invalidargument;
  return 1;
}  /* do_unrecognized */


int do_usage (OPTION*, const char*)
{
  printf (
"%s version %s\n"
"usage: %s [options] <file> [ [options] <file> ... ]\n"
"  -c, --checksignature Check signature (*)\n"
"  -H, --hash		Rewrite files with hash only\n"
"  -s, --sign           Rewrite files with hash and signature\n"
"  -o, --output FILE    Save new version to FILE\n"
"  -q, --quiet          Inhibit messages, call must rely on exit status\n"
"  -d, --debug          Show debug messages\n"
"  -V, --version        Display version and copyright\n"
"  -h, --help           Usage message\n"
"                   (*) Default option\n"
	,
	  g_szApplication, g_szVersion,
	  g_szApplication);

  return 1;
}  /* do_usage */


int do_version (OPTION*, const char*)
{
  printf ("%s\n", g_szVersion);
  return 1;
}  /* do_version */


/* main

   entry point.  Note that this main is simplified to an idiom because
   of the options code we use.  Files are processed when non-option
   command line arguments are found.  Refer to OPTION_F_NOPTION.

*/

int main (int argc, char** argv)
{
  g_szApplication = parse_application (*argv);

  int argc_used;
  int result = parse_options (argc - 1, argv + 1, rgOptions, &argc_used);

  if (result) {
    switch (result) {
      case 1:
	break;
      default:
	printf ("parse error %d at word %d\n", result, argc_used + 1);
	break;
    }  /* switch */
    g_exitStatus = invalidargument;
    return 1;
  }  /* if */
  
  if (!g_fProcessedFile)
    do_usage (NULL, NULL);

  return g_exitStatus;
}  /* main */
